/*Bibliothque classique
 */
#include <stdio.h>
#include <stdlib.h>

/* Type de structures utilises pour les smaphores
	et la mmoire partages */
#include <sys/types.h>

/* Fonction pour la communication interprocessus
 */
#include <sys/ipc.h>

/* Fonction pour les smaphores
 */
#include <sys/sem.h>

/* Fontion pour la mmoire partage
 */
#include <sys/shm.h>

/* Fonction pour grer les messages d'erreurs
 */
#include <errno.h>

/* Fontion sur les chaines de caractre
 */
#include <string.h>


/*! creer_semaphore()
\return identifiant du smaphore
*/
int creer_semaphore(){

	key_t key;
	int semid;

	/*! ftok cre une cl unique
	\param tabdiscussion.tmp nom du fichier correspondant
	\param 1 version du projet
	\return key, cl unique
	*/
	if((key=ftok("tabdiscussion.tmp", 1))==-1){
		perror("ftok");
		exit(1);
	}

	/*! semget() cre le smaphore
	\param key cl d'identification du smaphore
	\param 1 nombre de smaphore
	\param IPC_CREAT option pour crer le smaphore
	\param droit droits de type {-rw-------}
	\return
	*/
	semid=semget(key,1,IPC_CREAT|0666);

	/*! semctl effectue des oprations de controle sur le smaphore
	\param semid identifiant du smaphore
	\param 0 numro du smaphore concern
	\param Setval option pour attribuer une valeur au smaphore
	\param 1 valeur du smaphore
	\return
	*/
	semctl(semid,0,SETVAL,1);

	return semid;
}

/*! prendre_semaphore() prend un smaphore
\param semid identifiant du smaphore
\return
*/
void prendre_semaphore(int semid){

	/* Structure contenant l'opration  ffectuer sur le smaphore
	 */
	struct sembuf operation;

	/*! semop() ralise les oprations sur les smaphores
	\param semid identifiant du smaphore
	\param operation sem_num = numro du smaphore concern par l'opration, sem_op = -1 pour prendre, sem_flg = SEM_UNDO pour librer le smaphore suite  la fermeture du processus
	\param Nb_Sem nombre d'opration  effectuer
	\return
	*/
	operation.sem_num=0;
	operation.sem_op=-1;
	operation.sem_flg=SEM_UNDO;
	semop(semid,&operation,1);
}

/*! prendre_semaphore() prend un smaphore
\param semid identifiant du smaphore
\return
*/
void vendre_semaphore(int semid){

	/* Structure contenant l'opration  ffectuer sur le smaphore
	 */
	struct sembuf operation;

	/*! semop() ralise les oprations sur les smaphores
	\param semid identifiant du smaphore
	\param operation sem_num = numro du smaphore concern par l'opration, sem_op = +1 pour vendre, sem_flg = SEM_UNDO pour librer le smaphore suite  la fermeture du processus
	\param Nb_Sem nombre d'opration  effectuer
	\return
	*/
	operation.sem_num=0;
	operation.sem_op=+1;
	operation.sem_flg=SEM_UNDO;
	semop(semid,&operation,1);
}

void supprimer_semaphore(int semid){
	semctl(semid,0,IPC_RMID);
}

/*! initialisation_memoire() cre une zone mmoire pour 2000 caractres
\return identifiant de la zone mmoire
*/
int initialisation_memoire(){

	int shmid;
	key_t key;

	/*! ftok cr une cl unique
	\param tabdiscussion.tmp nom du fichier correspondant
	\param 1 version du projet
	\return key, cl unique
	*/
	if((key=ftok("tabdiscussion.tmp", 1))==-1){
		perror("ftok");
		exit(1);
	}
	/*! shmget recherche si le segment memoire existe deja
	\param key la cle qui identifie notre segment memoire
	\param size la taille du segment memoire, on veut 200*10 caracteres
	\param 0600 recherche en fonction des droits sur la memoire {-rw-------}
	\return shmid identifiant du segment memoire
	*/
	if((shmid = shmget(key,2000*sizeof(char),0600))==-1){
		
		/* si le segment n'existe pas, on le cr
		 */

	 	/*! shmget creation du segment de memoire
  	 	\param key la cle qui identifie notre segment memoire
		\param size la taille du segment memoire, on veut 2000 caracteres
		\param IPC_CREAT|0600 creation de la memoire avec les droits -rw-------
		\return shmid identifiant du segment memoire
		*/
		if ((shmid = shmget (key, 2000*sizeof(char), IPC_CREAT | 0600))==-1){
			fprintf(stderr, 
			"\nImpossible de cr�r le segment de m�oire partag� (errno=%d)\n", 
			errno);
		}
	}

	return shmid;
}

/*! initialisation_discussion() permet d'accder  la mmoire sous la forme d'un tableau
\param shmid identifiant de la zone mmoire
\return pointeur vers le premier caractre du tableau
*/
char* initialisation_discussion(int shmid){

	char* tab_discussion;

	/*! shmat relie le segment au processus en cours
	\param shmid l'identifiant de la memoire
	\param NULL adresse de debut de la memoire, on prend la premiere valeur
	\param SHM_RND arrondi a une page pret la taille de la memoire
	\return memoire pointeur vers le debut de la zone memoire
	*/
	if((tab_discussion = shmat(shmid,NULL,SHM_RND))==NULL){
		perror("shmat");
		exit(1);
	}

	return tab_discussion;
}

/*! saisie_message() demande la saisie d'un message et l'enregistre dans la mmoire partage
\param semid identifiant du smaphore protgeant l'accs  la mmoire
\param tab_discussion tableau de caractre contenant les derniers messages envoys
\param message pointeur vers la chaine de caractres (tableau)  ajouter dans le tableau de discussion
\return void
*/
void inserer_message(int semid, char* tab_discussion, char* message){

	/* Ajouter le sparateur de messages
	 */
	strcat(message,";;");

	/*Dbut de la section critique
	 */
	prendre_semaphore(semid);

	/*! strcat concatne 2 chaines de caractres
	\param tab_discussion tableau de caractre contenant les derniers messages envoys
	\param message message saisi
	\return tab_discussion + message
	*/
	strcat(tab_discussion,message);

	vendre_semaphore(semid);
	/* Fin section critique
	 */
}

/*! creer_memoire() crer la zone de mmoire partage
\param semid identifiant du smaphore protgeant l'accs  la mmoire
\param shmid identifiant de la zone mmoire partage
\param memoire poiteur vers le tableau de caractre contenant les derniers messages envoys
\return void
*/
void creer_memoire(int* semid, int* shmid, char** memoire){

	/* Crer le smaphore
	 */
	*semid = creer_semaphore();

	/* Crer le segment de mmoire partage
	 */
	*shmid = initialisation_memoire();

	/* Recuprer le segment de mmoire pour le tableau de la discussion
	 */
	*memoire = initialisation_discussion(*shmid);
}

/*! detruire_memoire() supprime la zone de mmoire partage
\param semid identifiant du smaphore protgeant l'accs  la mmoire
\param shmid identifiant de la zone mmoire partage
\param memoire poiteur vers le tableau de caractre contenant les derniers messages envoys
\return void
*/
void detruire_memoire(int semid, int shmid, char* memoire){

	/*! shmdt detache le segment memoire du processus
	\param memoire pointeur vers le debut de la zone memoire
	*/
	shmdt(memoire);
	
	/*! shmctl controle de la zone memoire
	\param shmid identifiant de la zone memoire
	\param IPC_RMID action de suppression de la zone mmoire
	\param NULL adresse de debut de la memoire
	*/
	shmctl(shmid,IPC_RMID,NULL);

	/* Supprimer le smaphore
	 */
	supprimer_semaphore(semid);
}
